#!usr/bin/env python3
# -*- coding:utf-8 -*-

__author__ = 'mayanqiong'

import asyncio
import time
from abc import abstractmethod
from typing import Type, Union

from tqsdk.channel import TqChan
from tqsdk.datetime import _get_trading_day_from_timestamp, _get_trading_day_end_time, _get_trade_timestamp, \
    _is_in_trading_time, _timestamp_nano_to_str, _str_to_timestamp_nano
from tqsdk.diff import _get_obj, _register_update_chan, _merge_diff
from tqsdk.entity import Entity
from tqsdk.objs import Quote
from tqsdk.tradeable.tradeable import Tradeable
from tqsdk.tradeable.sim.trade_future import SimTrade
from tqsdk.tradeable.sim.trade_stock import SimTradeStock
from tqsdk.utils import _query_for_quote


class BaseSim(Tradeable):

    def __init__(self, account_id, init_balance, trade_class: Union[Type[SimTrade], Type[SimTradeStock]]) -> None:
        self._account_id = account_id
        super(BaseSim, self).__init__()

        self.trade_log = {}  # 日期->交易记录及收盘时的权益及持仓
        self.tqsdk_stat = {}  # 回测结束后储存回测报告信息
        self._init_balance = init_balance
        self._current_datetime = "1990-01-01 00:00:00.000000"  # 当前行情时间（最新的 quote 时间）
        self._trading_day_end = "1990-01-01 18:00:00.000000"
        self._local_time_record = float("nan")  # 记录获取最新行情时的本地时间
        self._sim_trade = trade_class(account_key=self._account_key,
                                      account_id=self._account_id,
                                      init_balance=self._init_balance,
                                      get_trade_timestamp=self._get_trade_timestamp,
                                      is_in_trading_time=self._is_in_trading_time)
        self._data = Entity()
        self._data._instance_entity([])
        self._prototype = {
            "quotes": {
                "#": Quote(self),  # 行情的数据原型
            }
        }
        self._quote_tasks = {}

    @property
    def _account_name(self):
        return self._account_id

    @property
    def _account_info(self):
        info = super(BaseSim, self)._account_info
        info.update({
            "account_id": self._account_id
        })
        return info

    async def _run(self, api, api_send_chan, api_recv_chan, md_send_chan, md_recv_chan):
        """模拟交易task"""
        self._api = api
        self._tqsdk_backtest = {}  # 储存可能的回测信息
        self._logger = api._logger.getChild("TqSim")  # 调试信息输出
        self._api_send_chan = api_send_chan
        self._api_recv_chan = api_recv_chan
        self._md_send_chan = md_send_chan
        self._md_recv_chan = md_recv_chan
        # True 下游发过 subscribe，但是没有转发给上游；False 表示下游发的 subscribe 都转发给上游
        self._pending_subscribe_downstream = False
        # True 发给上游 subscribe，但是没有收到过回复；False 如果行情不变，上游不会回任何包
        self._pending_subscribe_upstream = False
        self._all_subscribe = set()  # 客户端+模拟交易模块订阅的合约集合
        # 是否已经发送初始账户信息
        self._has_send_init_account = False
        try:
            await super(BaseSim, self)._run(api, api_send_chan, api_recv_chan, md_send_chan, md_recv_chan)
        finally:
            self._handle_stat_report()
            for s in self._quote_tasks:
                self._quote_tasks[s]["task"].cancel()
            await asyncio.gather(*[self._quote_tasks[s]["task"] for s in self._quote_tasks], return_exceptions=True)

    async def _handle_recv_data(self, pack, chan):
        """
        处理所有上游收到的数据包，这里应该将需要发送给下游的数据 append 到 self._diffs
        pack: 收到的数据包
        chan: 收到此数据包的 channel
        """
        self._pending_subscribe_upstream = False
        if pack["aid"] == "rtn_data":
            self._md_recv(pack)  # md_recv 中会发送 wait_count 个 quotes 包给各个 quote_chan
            await asyncio.gather(*[quote_task["quote_chan"].join() for quote_task in self._quote_tasks.values()])
        if self._tqsdk_backtest != {} and self._tqsdk_backtest["current_dt"] >= self._tqsdk_backtest["end_dt"]:
            # 回测情况下，把 _handle_stat_report 在循环中回测结束时执行
            self._handle_stat_report()

    async def _handle_req_data(self, pack):
        """
        处理所有下游发送的非 peek_message 数据包
        这里应该将发送的请求转发到指定的某个上游 channel
        """
        if self._is_self_trade_pack(pack):
            if pack["aid"] == "insert_order":
                symbol = pack["exchange_id"] + "." + pack["instrument_id"]
                if symbol not in self._quote_tasks:
                    quote_chan = TqChan(self._api)
                    order_chan = TqChan(self._api)
                    self._quote_tasks[symbol] = {
                        "quote_chan": quote_chan,
                        "order_chan": order_chan,
                        "task": self._api.create_task(self._quote_handler(symbol, quote_chan, order_chan))
                    }
                await self._quote_tasks[symbol]["order_chan"].send(pack)
            else:
                # pack 里只有 order_id 信息，发送到每一个合约的 order_chan, 交由 quote_task 判断是不是当前合约下的委托单
                for symbol in self._quote_tasks:
                    await self._quote_tasks[symbol]["order_chan"].send(pack)
        elif pack["aid"] == "subscribe_quote":
            # 这里只会增加订阅合约，不会退订合约
            await self._subscribe_quote(set(pack["ins_list"].split(",")))
        else:
            await self._md_send_chan.send(pack)

    async def _on_send_diff(self, pending_peek):
        if pending_peek and self._pending_subscribe_downstream:
            await self._send_subscribe_quote()

    async def _subscribe_quote(self, symbols: [set, str]):
        """
        这里只会增加订阅合约，不会退订合约
        todo: 这里用到了 self._pending_peek ，父类的内部变量
        """
        symbols = symbols if isinstance(symbols, set) else {symbols}
        if symbols - self._all_subscribe:
            self._all_subscribe |= symbols
            if self._pending_peek and not self._pending_subscribe_upstream:
                await self._send_subscribe_quote()
            else:
                self._pending_subscribe_downstream = True

    async def _send_subscribe_quote(self):
        self._pending_subscribe_upstream = True
        self._pending_subscribe_downstream = False
        await self._md_send_chan.send({
            "aid": "subscribe_quote",
            "ins_list": ",".join(self._all_subscribe)
        })

    def _handle_stat_report(self):
        if self.tqsdk_stat:
            return
        self._settle()
        self._report()
        self._diffs.append({
            "trade": {
                self._account_key: {
                    "accounts": {
                        "CNY": {
                            "_tqsdk_stat": self.tqsdk_stat
                        }
                    }
                }
            }
        })

    async def _ensure_quote_info(self, symbol, quote_chan):
        """quote收到合约信息后返回"""
        quote = _get_obj(self._data, ["quotes", symbol], Quote(self._api))
        if quote.get("price_tick") == quote.get("price_tick"):
            return quote.copy()
        if quote.get("price_tick") != quote.get("price_tick"):
            await self._md_send_chan.send(_query_for_quote(symbol))
        async for _ in quote_chan:
            quote_chan.task_done()
            if quote.get("price_tick") == quote.get("price_tick"):
                return quote.copy()

    async def _ensure_quote(self, symbol, quote_chan):
        """quote收到行情以及合约信息后返回"""
        quote = _get_obj(self._data, ["quotes", symbol], Quote(self._api))
        _register_update_chan(quote, quote_chan)
        if quote.get("datetime", "") and quote.get("price_tick") == quote.get("price_tick"):
            return quote.copy()
        if quote.get("price_tick") != quote.get("price_tick"):
            # 对于没有合约信息的 quote，发送查询合约信息的请求
            await self._md_send_chan.send(_query_for_quote(symbol))
        async for _ in quote_chan:
            quote_chan.task_done()
            if quote.get("datetime", "") and quote.get("price_tick") == quote.get("price_tick"):
                return quote.copy()

    async def _quote_handler(self, symbol, quote_chan, order_chan):
        try:
            await self._subscribe_quote(symbol)
            quote = await self._ensure_quote(symbol, quote_chan)
            if quote["ins_class"].endswith("INDEX") and quote["exchange_id"] == "KQ":
                # 指数可以交易，需要补充 margin commission
                if "margin" not in quote:
                    quote_m = await self._ensure_quote_info(symbol.replace("KQ.i", "KQ.m"), quote_chan)
                    quote_underlying = await self._ensure_quote_info(quote_m["underlying_symbol"], quote_chan)
                    self._data["quotes"][symbol]["margin"] = quote_underlying["margin"]
                    self._data["quotes"][symbol]["commission"] = quote_underlying["commission"]
                    quote.update(self._data["quotes"][symbol])
            underlying_quote = None
            if quote["ins_class"].endswith("OPTION"):
                # 如果是期权，订阅标的合约行情，确定收到期权标的合约行情
                underlying_symbol = quote["underlying_symbol"]
                await self._subscribe_quote(underlying_symbol)
                underlying_quote = await self._ensure_quote(underlying_symbol, quote_chan)  # 订阅合约
            # 在等待标的行情的过程中，quote_chan 可能有期权行情，把 quote_chan 清空，并用最新行情更新 quote
            while not quote_chan.empty():
                quote_chan.recv_nowait()
                quote_chan.task_done()

            # 用最新行情更新 quote
            quote.update(self._data["quotes"][symbol])
            if underlying_quote:
                underlying_quote.update(self._data["quotes"][underlying_symbol])
            task = self._api.create_task(self._forward_chan_handler(order_chan, quote_chan))
            quotes = {symbol: quote}
            if underlying_quote:
                quotes[underlying_symbol] = underlying_quote
            self._sim_trade.update_quotes(symbol, {"quotes": quotes})
            async for pack in quote_chan:
                if "aid" not in pack:
                    diffs, orders_events = self._sim_trade.update_quotes(symbol, pack)
                    self._handle_diffs(diffs, orders_events, "match order")
                elif pack["aid"] == "insert_order":
                    diffs, orders_events = self._sim_trade.insert_order(symbol, pack)
                    self._handle_diffs(diffs, orders_events, "insert order")
                elif pack["aid"] == "cancel_order":
                    diffs, orders_events = self._sim_trade.cancel_order(symbol, pack)
                    self._handle_diffs(diffs, orders_events, "cancel order")
                quote_chan.task_done()
        finally:
            await quote_chan.close()
            await order_chan.close()
            task.cancel()
            await asyncio.gather(task, return_exceptions=True)

    async def _forward_chan_handler(self, chan_from, chan_to):
        async for pack in chan_from:
            await chan_to.send(pack)

    def _md_recv(self, pack):
        for d in pack["data"]:
            self._diffs.append(d)
            # 在第一次收到 mdhis_more_data 为 False 的时候，发送账户初始截面信息，这样回测模式下，往后的模块才有正确的时间顺序
            if not self._has_send_init_account and not d.get("mdhis_more_data", True):
                self._diffs.append(self._sim_trade.init_snapshot())
                self._diffs.append({
                    "trade": {
                        self._account_key: {
                            "trade_more_data": False
                        }
                    }
                })
                self._has_send_init_account = True
            _tqsdk_backtest = d.get("_tqsdk_backtest", {})
            if _tqsdk_backtest:
                # 回测时，用 _tqsdk_backtest 对象中 current_dt 作为 TqSim 的 _current_datetime
                self._tqsdk_backtest.update(_tqsdk_backtest)
                self._current_datetime = _timestamp_nano_to_str(self._tqsdk_backtest["current_dt"])
                self._local_time_record = float("nan")
                # 1. 回测时不使用时间差来模拟交易所时间的原因(_local_time_record始终为初始值nan)：
                #   在sim收到行情后记录_local_time_record，然后下发行情到api进行merge_diff(),api需要处理完k线和quote才能结束wait_update(),
                #   若处理时间过长，此时下单则在判断下单时间时与测试用例中的预期时间相差较大，导致测试用例无法通过。
                # 2. 回测不使用时间差的方法来判断下单时间仍是可行的: 与使用了时间差的方法相比, 只对在每个交易时间段最后一笔行情时的下单时间判断有差异,
                #   若不使用时间差, 则在最后一笔行情时下单仍判断为在可交易时间段内, 且可成交.
            quotes_diff = d.get("quotes", {})
            # 先根据 quotes_diff 里的 datetime, 确定出 _current_datetime，再 _merge_diff(同时会发送行情到 quote_chan)
            for symbol, quote_diff in quotes_diff.items():
                if quote_diff is None:
                    continue
                # 若直接使用本地时间来判断下单时间是否在可交易时间段内 可能有较大误差,因此判断的方案为:(在接收到下单指令时判断 估计的交易所时间 是否在交易时间段内)
                # 在更新最新行情时间(即self._current_datetime)时，记录当前本地时间(self._local_time_record)，
                # 在这之后若收到下单指令，则获取当前本地时间,判 "最新行情时间 + (当前本地时间 - 记录的本地时间)" 是否在交易时间段内。
                # 另外, 若在盘后下单且下单前未订阅此合约：
                # 因为从_md_recv()中获取数据后立即判断下单时间则速度过快(两次time.time()的时间差小于最后一笔行情(14:59:9995)到15点的时间差),
                # 则会立即成交,为处理此情况则将当前时间减去5毫秒（模拟发生5毫秒网络延迟，则两次time.time()的时间差增加了5毫秒）。
                # todo: 按交易所来存储 _current_datetime(issue： #277)
                if quote_diff.get("datetime", "") > self._current_datetime:
                    # 回测时，当前时间更新即可以由 quote 行情更新，也可以由 _tqsdk_backtest.current_dt 更新，
                    # 在最外层的循环里，_tqsdk_backtest.current_dt 是在 rtn_data.data 中数组位置中的最后一个，会在循环最后一个才更新 self.current_datetime
                    # 导致前面处理 order 时的 _current_datetime 还是旧的行情时间
                    self._current_datetime = quote_diff["datetime"]  # 最新行情时间
                    # 更新最新行情时间时的本地时间，回测时不使用时间差
                    self._local_time_record = (time.time() - 0.005) if not self._tqsdk_backtest else float("nan")
                if self._current_datetime > self._trading_day_end:  # 结算
                    self._settle()
                    # 若当前行情时间大于交易日的结束时间(切换交易日)，则根据此行情时间更新交易日及交易日结束时间
                    trading_day = _get_trading_day_from_timestamp(self._get_current_timestamp())
                    self._trading_day_end = _timestamp_nano_to_str(_get_trading_day_end_time(trading_day) - 999)
            if quotes_diff:
                _merge_diff(self._data, {"quotes": quotes_diff}, self._prototype, persist=False, reduce_diff=False, notify_update_diff=True)

    def _handle_diffs(self, diffs, orders_events, msg):
        """
        处理 sim_trade 返回的 diffs
        orders_events 为持仓变更事件，依次屏幕输出信息，打印日志
        """
        self._diffs += diffs
        for order in orders_events:
            if order["status"] == "FINISHED":
                self._handle_on_finished(msg, order)
            else:
                assert order["status"] == "ALIVE"
                self._handle_on_alive(msg, order)

    def _settle(self):
        if self._trading_day_end[:10] == "1990-01-01":
            return
        # 结算并记录账户截面
        diffs, orders_events, trade_log = self._sim_trade.settle()
        self._handle_diffs(diffs, orders_events, "settle")
        self.trade_log[self._trading_day_end[:10]] = trade_log

    @abstractmethod
    def _handle_on_alive(self, msg, order):
        """
        在 order 状态变为 ALIVE 调用，屏幕输出信息，打印日志
        """
        pass

    @abstractmethod
    def _handle_on_finished(self, msg, order):
        """
        在 order 状态变为 FINISHED 调用，屏幕输出信息，打印日志
        """
        pass

    @abstractmethod
    def _report(self):
        pass

    def _get_current_timestamp(self):
        return _str_to_timestamp_nano(self._current_datetime)

    def _get_trade_timestamp(self):
        return _get_trade_timestamp(self._current_datetime, self._local_time_record)

    def _is_in_trading_time(self, quote):
        return _is_in_trading_time(quote, self._current_datetime, self._local_time_record)



#!usr/bin/env python3
# -*- coding:utf-8 -*-
__author__ = 'yanqiong'

from typing import Optional

from tqsdk.tradeable.otg.base_otg import BaseOtg
from tqsdk.tradeable.mixin import FutureMixin, StockMixin


class TqKq(BaseOtg, FutureMixin):
    """天勤快期模拟账户类"""

    def __init__(self, td_url: Optional[str] = None, number: Optional[int] = None):
        """
        创建快期模拟账户实例

        快期模拟的账户和交易信息可以在快期专业版查看，可以点击 `快期专业版 <https://www.shinnytech.com/qpro/>`_ 进行下载

        Args:
            td_url (str): [可选]指定交易服务器的地址, 默认使用快期账户对应的交易服务地址

            number (int): [可选]模拟交易账号编号, 默认为主模拟账号, 可以通过指定 1~99 的数字来使用辅模拟帐号, 各个帐号的数据完全独立, 使用该功能需要购买专业版的权限, 且对应的辅账户可以在快期专业版上登录并进行观察

        Example1::

            from tqsdk import TqApi, TqAuth, TqKq

            tq_kq = TqKq()
            api = TqApi(account=tq_kq, auth=TqAuth("快期账户", "账户密码"))
            quote = api.get_quote("SHFE.cu2206")
            print(quote)
            # 下单限价单
            order = api.insert_order(symbol="SHFE.cu2206", direction='BUY', offset='OPEN', limit_price=quote.last_price, volume=1)
            while order.status == 'ALIVE':
                api.wait_update()
                print(order)  # 打印委托单信息

            print(tq_kq.get_account())  # 打印快期模拟账户信息

            print(tq_kq.get_position("SHFE.cu2206"))  # 打印持仓信息

            for trade in order.trade_records.values():
                print(trade)  # 打印委托单对应的成交信息
            api.close()

        Example2::

            from tqsdk import TqApi, TqAuth, TqKq, TqMultiAccount

            # 创建快期模拟账户和辅模拟账户001
            tq_kq = TqKq()
            tq_kq001= TqKq(number=1)

            # 使用多账户模块同时登录这两个模拟账户
            api = TqApi(account=TqMultiAccount([tq_kq,tq_kq001]), auth=TqAuth("快期账户", "账户密码"))

            print(tq_kq.get_account())  # 打印快期模拟账户信息

            print(tq_kq001.get_account())  # 打印快期模拟001账户信息

            api.close()


        """
        super().__init__("快期模拟", str(number) if number else "", "", td_url=td_url)
        self._account_no = number

    @property
    def _account_name(self):
        # 用于界面展示的用户信息
        if self._account_no:
            return f'{self._api._auth._user_name}:{self._account_no:03d}'
        else:
            return self._api._auth._user_name

    @property
    def _account_info(self):
        info = super(TqKq, self)._account_info
        info.update({
            "account_type": self._account_type
        })
        return info

    @property
    def _account_auth(self):
        return {
            "feature": "tq_ma" if self._account_no else None,
            "account_id": None,
            "auto_add": False,
        }

    def _update_otg_info(self, api):
        self._account_id = f'{api._auth._auth_id}{self._account_no:03d}' if self._account_no else api._auth._auth_id
        self._password = f'shinnytech{self._account_no:03d}' if self._account_no else api._auth._auth_id
        super(TqKq, self)._update_otg_info(api)


class TqKqStock(BaseOtg, StockMixin):
    """天勤实盘类"""

    def __init__(self, td_url: Optional[str] = None, number: Optional[int] = None):
        """
        创建快期股票模拟账户实例

        快期股票模拟为专业版功能，可以点击 `天勤量化专业版 <https://www.shinnytech.com/tqsdk_professional/>`_ 申请试用或购买

        Args:
            td_url (str): [可选]指定交易服务器的地址, 默认使用快期账户对应的交易服务地址

            number (int): [可选]模拟交易账号编号, 默认为主模拟账号, 可以通过指定 1~99 的数字来使用辅模拟帐号, 各个帐号的数据完全独立

        Example::

            from tqsdk import TqApi, TqAuth, TqKqStock, TqChan

            tq_kq_stock = TqKqStock()
            api = TqApi(account=tq_kq_stock, auth=TqAuth("快期账户", "账户密码"))
            quote = api.get_quote("SSE.688529")
            print(quote)
            # 下单限价单
            order = api.insert_order("SSE.688529", volume=200, direction="BUY", limit_price=quote.ask_price1)
            while order.status == 'ALIVE':
                api.wait_update()
                print(order)  # 打印委托单信息

            print(tq_kq_stock.get_account())  # 打印快期股票模拟账户信息

            print(tq_kq_stock.get_position("SSE.688529"))  # 打印持仓信息

            for trade in order.trade_records.values():
                print(trade)  # 打印委托单对应的成交信息
            api.close()

        """
        super().__init__("快期股票模拟", str(number) if number else "", "", td_url=td_url)
        self._account_no = number

    @property
    def _account_name(self):
        # 用于界面展示的用户信息
        if self._account_no:
            return f'{self._api._auth._user_name}_stock:{self._account_no:03d}'
        else:
            return self._api._auth._user_name + "_stock"

    @property
    def _account_info(self):
        info = super(TqKqStock, self)._account_info
        info.update({
            "account_type": self._account_type
        })
        return info

    @property
    def _account_auth(self):
        return {
            "feature": "tq_ma" if self._account_no else None,
            "account_id": self._auth_account_id,
            "auto_add": False,
        }

    def _update_otg_info(self, api):
        self._auth_account_id = api._auth._auth_id + "-sim-securities"
        self._account_id = f'{api._auth._auth_id}{self._account_no:03d}-sim-securities' if self._account_no else self._auth_account_id
        self._password = f'shinnytech{self._account_no:03d}' if self._account_no else api._auth._auth_id
        super(TqKqStock, self)._update_otg_info(api)



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'mayanqiong'

import numpy as np
from pandas import DataFrame

from tqsdk.datetime import _get_trading_timestamp, _get_trade_timestamp
from tqsdk.rangeset import _rangeset_head, _rangeset_slice, _rangeset_length

"""
检查参数类型
"""

from inspect import isfunction


def _check_volume_limit(min_volume, max_volume):
    if min_volume is not None and min_volume <= 0:
        raise Exception("最小下单手数(min_volume) %s 错误, 请检查 min_volume 是否填写正确" % min_volume)
    if max_volume is not None and max_volume <= 0:
        raise Exception("最大下单手数(max_volume) %s 错误, 请检查 max_volume 是否填写正确" % max_volume)
    if (min_volume is None and max_volume) or (max_volume is None and min_volume):
        raise Exception("最小下单手数(min_volume) %s 和 最大下单手数(max_volume) %s 必须用时填写" % (min_volume, max_volume))
    if min_volume and max_volume and min_volume > max_volume:
        raise Exception("最小下单手数(min_volume) %s ，最大下单手数(max_volume) %s 错误, 请检查 min_volume, max_volume 是否填写正确" % (
            min_volume, max_volume))
    return int(min_volume) if min_volume else None, int(max_volume) if max_volume else None


def _check_direction(direction):
    if direction not in ("BUY", "SELL"):
        raise Exception("下单方向(direction) %s 错误, 请检查 direction 参数是否填写正确" % direction)
    return direction


def _check_offset(offset):
    if offset not in ("OPEN", "CLOSE", "CLOSETODAY"):
        raise Exception("开平标志(offset) %s 错误, 请检查 offset 是否填写正确" % offset)
    return offset


def _check_offset_priority(offset_priority):
    if len(offset_priority.replace(",", "").replace("今", "", 1).replace("昨", "", 1).replace("开", "", 1)) > 0:
        raise Exception("开平仓顺序(offset_priority) %s 错误, 请检查 offset_priority 参数是否填写正确" % offset_priority)
    return offset_priority


def _check_volume(volume):
    _volume = int(volume)
    if _volume <= 0:
        raise Exception("下单手数(volume) %s 错误, 请检查 volume 是否填写正确" % volume)
    return _volume


def _check_price(price):
    if price in ("ACTIVE", "PASSIVE") or isfunction(price):
        return price
    else:
        raise Exception("下单方式(price) %s 错误, 请检查 price 参数是否填写正确" % price)


def _check_time_table(time_table: DataFrame):
    if not isinstance(time_table, DataFrame):
        raise Exception(f"time_table 参数应该是 pandas.DataFrame 类型")
    need_columns = {'price', 'target_pos', 'interval'} - set(time_table.columns)
    if need_columns:
        raise Exception(f"缺少必要的列 {need_columns}")
    if time_table.shape[0] > 0:
        if time_table['interval'].isnull().values.any() or np.where(time_table['interval'] < 0, True, False).any():
            raise Exception(f"interval 列必须为正数，请检查参数 {time_table['interval']}")
        if time_table['target_pos'].isnull().values.any() or not np.issubdtype(time_table['target_pos'].dtype, np.integer):
            raise Exception(f"target_pos 列必须为整数，请检查参数 {time_table['target_pos']}")
        if not (np.isin(time_table['price'], ('PASSIVE', 'ACTIVE', None)) | time_table['price'].apply(isfunction)).all():
            raise Exception(f"price 列必须为 ('PASSIVE', 'ACTIVE', None, Callable) 之一，请检查参数 {time_table['price']}")
    return time_table


def _get_deadline_from_interval(quote, interval):
    """将 interval （持续长度 seconds）列转换为 deadline（结束时间 nano_timestamp）"""
    # 当前交易日完整的交易时间段
    trading_timestamp = _get_trading_timestamp(quote, quote.datetime)
    trading_timestamp_nano_range = trading_timestamp['night'] + trading_timestamp['day']  # 当前交易日完整的交易时间段
    # 当前时间 行情时间
    current_timestamp_nano = _get_trade_timestamp(quote.datetime, float('nan'))
    if not trading_timestamp_nano_range[0][0] <= current_timestamp_nano < trading_timestamp_nano_range[-1][1]:
        raise Exception("当前时间不在指定的交易时间段内")
    deadline = []
    for index, value in interval.items():
        r = _rangeset_head(_rangeset_slice(trading_timestamp_nano_range, current_timestamp_nano), int(value * 1e9))
        if _rangeset_length(r) < int(value * 1e9):
            raise Exception("指定时间段超出当前交易日")
        deadline.append(r[-1][1])
        current_timestamp_nano = r[-1][1]
    return deadline



#!usr/bin/env python3
# -*- coding:utf-8 -*-

__author__ = 'mayanqiong'

from typing import Optional
from abc import ABC, abstractmethod

from tqsdk.baseModule import TqModule


class Tradeable(ABC, TqModule):

    def __init__(self):
        self._account_key = self._get_account_key()  # 每个账户的唯一标识，在账户初始化时就确定下来，后续只读不写

    def _get_account_key(self):
        return str(id(self))

    @property
    @abstractmethod
    def _account_name(self):
        # 用于界面展示的用户信息
        raise NotImplementedError

    @property
    def _account_info(self):
        # 用于 web_helper 获取初始账户信息
        return {
            "account_key": self._account_key,
            "account_name": self._account_name
        }

    @property
    def _account_auth(self):
        # 使用该账户需要的授权信息
        return {
            "feature": None,
            "account_id": None,
            "auto_add": False,
        }

    def _is_self_trade_pack(self, pack):
        """是否是当前交易实例应该处理的交易包"""
        if pack["aid"] in ["insert_order", "cancel_order", "set_risk_management_rule"]:
            assert "account_key" in pack, "发给交易请求的包必须包含 account_key"
            if pack["account_key"] != self._account_key:
                return False
            else:
                pack.pop("account_key", None)
                return True
        return False

    def _connect_td(self, api, index: int) -> Optional[str]:
        # 用于建立交易连接
        return None



from tqsdk import TqApi, TqAuth

'''
获取标的对应看涨期权的期权和行权价对应列表
'''
api = TqApi(auth=TqAuth("快期账户", "账号密码"))

# 获取沪深300股指期权的认购在市合约
ls = api.query_options("SSE.000300", "CALL", expired=False)

# 批量获取这些合约的quote合约信息
quote_ls = api.get_quote_list(ls)

option_ls = {}

# 遍历quote合约信息，将合约和其对应行权价组装成字典
for i in quote_ls:
    option_ls[i.instrument_id] = i.strike_price

print(option_ls)

api.close()


#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'limin'

'''
双均线策略
注: 该示例策略仅用于功能示范, 实盘时请根据自己的策略/经验进行修改
'''
from tqsdk import TqApi, TqAuth, TargetPosTask
from tqsdk.tafunc import ma

SHORT = 30  # 短周期
LONG = 60  # 长周期
SYMBOL = "SHFE.bu2012"  # 合约代码

api = TqApi(auth=TqAuth("快期账户", "账户密码"))
print("策略开始运行")

data_length = LONG + 2  # k线数据长度
# "duration_seconds=60"为一分钟线, 日线的duration_seconds参数为: 24*60*60
klines = api.get_kline_serial(SYMBOL, duration_seconds=60, data_length=data_length)
target_pos = TargetPosTask(api, SYMBOL)

while True:
    api.wait_update()

    if api.is_changing(klines.iloc[-1], "datetime"):  # 产生新k线:重新计算SMA
        short_avg = ma(klines["close"], SHORT)  # 短周期
        long_avg = ma(klines["close"], LONG)  # 长周期

        # 均线下穿，做空
        if long_avg.iloc[-2] < short_avg.iloc[-2] and long_avg.iloc[-1] > short_avg.iloc[-1]:
            target_pos.set_target_volume(-3)
            print("均线下穿，做空")

        # 均线上穿，做多
        if short_avg.iloc[-2] < long_avg.iloc[-2] and short_avg.iloc[-1] > long_avg.iloc[-1]:
            target_pos.set_target_volume(3)
            print("均线上穿，做多")



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'limin'

from tqsdk import TqApi, TqAuth

'''
画图示例: 在主图中画线和方框
注意: 画图示例中用到的数据不含有实际意义，请根据自己的实际策略情况进行修改
'''

api = TqApi(web_gui=True, auth=TqAuth("快期账户", "账户密码")) # web_gui=True, 开启使用 web 界面查看绘图结果的功能
klines = api.get_kline_serial("SHFE.rb2105", 60)

# 由于需要在浏览器中查看绘图结果，因此程序不能退出
while True:
    api.wait_update() # 当有业务信息发生变化时执行
    # 当最后 1 根柱子最大最小值价差大于 0.05 时，在主图绘制信号
    high = klines.iloc[-1].high
    low = klines.iloc[-1].low
    if high - low > 0.05:
        # 绘制直线, 每一个 id 对应同一条直线
        api.draw_line(klines, -1, high, -1, low, id="box%.0f" % (klines.iloc[-1].id), color=0xaa662244, width=4)
        # 绘制字符串
        api.draw_text(klines, "信号1", x=-1, y=low, id="text%.0f" % (klines.iloc[-1].id), color=0xFFFF3333)




# !usr/bin/env python3
# -*- coding:utf-8 -*-
__author__ = 'yanqiong'

import datetime
import os

import psutil

DEBUG_DIR = os.path.join(os.path.expanduser('~'), ".tqsdk/logs")


def _get_log_name():
    """返回默认 debug 文件生成的位置"""
    if not os.path.exists(DEBUG_DIR):
        os.makedirs(os.path.join(os.path.expanduser('~'), ".tqsdk/logs"), exist_ok=True)
    return os.path.join(DEBUG_DIR, f"{datetime.datetime.now().strftime('%Y%m%d%H%M%S%f')}-{os.getpid()}.log")


def _get_disk_free():
    free = psutil.disk_usage(DEBUG_DIR).free
    return free / 1e9


def _log_path_list():
    # 获取所有日志文件路径，并按照修改时间递减排序
    path_list = [os.path.join(DEBUG_DIR, log) for log in os.listdir(DEBUG_DIR)]
    path_list.sort(key=lambda x: _stat_dt(x))
    return path_list


def _stat_dt(path):
    try:
        return datetime.datetime.fromtimestamp(os.stat(path).st_mtime)
    except:
        return datetime.datetime.now()


def _remove_log(path):
    try:
        os.remove(path)
    except:
        pass  # 忽略抛错


def _clear_logs():
    """清除最后修改时间是 n 天前的日志"""
    if not os.path.exists(DEBUG_DIR):
        return
    n = os.getenv("TQ_SAVE_LOG_DAYS", 30)
    # 清除最后修改时间是 n 天前的日志
    # 清空日志保证剩余空间大于 3G，但是最近 3 个自然日的一定不会清除，保证最近的一个交易日不会被清除日志
    dt30 = datetime.datetime.now() - datetime.timedelta(days=int(n))
    dt3 = datetime.datetime.now() - datetime.timedelta(days=int(3))
    for path in _log_path_list():
        if _stat_dt(path) < dt30 or (_get_disk_free() < 3 and _stat_dt(path) < dt3):
            _remove_log(path)
        else:
            break



#!usr/bin/env python3
#-*- coding:utf-8 -*-
__author__ = 'yanqiong'

import copy
from typing import Set, Union, Dict, Tuple

from tqsdk.entity import Entity


def _merge_diff(result, diff, prototype, persist, reduce_diff=False, notify_update_diff=False):
    """
    更新业务数据,并同步发送更新通知，保证业务数据的更新和通知是原子操作
    :param result: 更新结果
    :param diff: diff pack
    :param persist: 是否保留 result 中所有字段，如果为 True，则 diff 里为 None 的对象在 result 中不会删除；如果是 False，则 result 结果中会删除 diff 里为 None 的对象
    :param reduce_diff: 表示是否修改 diff 对象本身，如果为 True 函数运行完成后，diff 会更新为与 result 真正的有区别的字段；如果为 False，diff 不会修改
        默认不会修改 diff，只有 api 中 is_changing 接口需要 diffs 为真正有变化的数据
    :param notify_update_diff: 为 True 表示发送更新通知的发送的是包含 diff 的完整数据包（方便 TqSim 中能每个合约的 task 可以单独维护自己的数据），反之只发送 True
    :return:
    """
    for key in list(diff.keys()):
        value_type = type(diff[key])
        if value_type is str and key in prototype and not type(prototype[key]) is str:
            diff[key] = prototype[key]
        if diff[key] is None:
            if (persist or "#" in prototype) and reduce_diff:
                del diff[key]
            else:
                if notify_update_diff:
                    dv = result.pop(key, None)
                    _notify_update(dv, True, _gen_diff_obj(None, result["_path"] + [key]))
                else:
                    dv = result.pop(key, None)
                    _notify_update(dv, True, True)
        elif value_type is dict or value_type is Entity:
            default = None
            tpersist = persist
            if key in prototype:
                tpt = prototype[key]
            elif "*" in prototype:
                tpt = prototype["*"]
            elif "@" in prototype:
                tpt = prototype["@"]
                default = tpt
            elif "#" in prototype:
                tpt = prototype["#"]
                default = tpt
                tpersist = True
            else:
                tpt = {}
            target = _get_obj(result, [key], default=default)
            _merge_diff(target, diff[key], tpt, persist=tpersist, reduce_diff=reduce_diff, notify_update_diff=notify_update_diff)
            if reduce_diff and len(diff[key]) == 0:
                del diff[key]
        elif reduce_diff and key in result and (
                result[key] == diff[key] or (diff[key] != diff[key] and result[key] != result[key])):
            # 判断 diff[key] != diff[key] and result[key] != result[key] 以处理 value 为 nan 的情况
            del diff[key]
        else:
            result[key] = diff[key]
    if len(diff) != 0:
        diff_obj = True
        if notify_update_diff:
            # 这里发的数据目前是不需要 copy (浅拷贝会有坑，深拷贝的话性能不知道有多大影响)
            # 因为这里现在会用到发送这个 diff 的只有 quote 对象，只有 sim 会收到使用，sim 收到之后是不会修改这个 diff
            # 所以这里就约定接收方不能改 diff 中的值
            diff_obj = _gen_diff_obj(diff, result["_path"])
        _notify_update(result, False, diff_obj)


def _gen_diff_obj(diff, path):
    """将 diff 根据 path 拼成一个完整的 diff 包"""
    diff_obj = diff
    for i in range(len(path)):
        diff_obj = {path[len(path)-i-1]: diff_obj}
    return diff_obj


def _notify_update(target, recursive, content):
    """同步通知业务数据更新"""
    if isinstance(target, dict) or isinstance(target, Entity):
        for q in getattr(target, "_listener", {}):
            q.send_nowait(content)
        if recursive:
            for v in target.values():
                _notify_update(v, recursive, content)


def _get_obj(root, path, default=None):
    """获取业务数据"""
    d = root
    for i in range(len(path)):
        if path[i] not in d:
            if i != len(path) - 1 or default is None:
                dv = Entity()
            else:
                dv = copy.copy(default)
            dv._instance_entity(d["_path"] + [path[i]])
            d[path[i]] = dv
        d = d[path[i]]
    return d


def _register_update_chan(objs, chan):
    if not isinstance(objs, list):
        objs = [objs]
    for o in objs:
        o["_listener"].add(chan)
    return chan


def _is_key_exist(diff, path, key):
    """判断指定数据是否存在"""
    for p in path:
        if not isinstance(diff, dict) or p not in diff:
            return False
        diff = diff[p]
    if not isinstance(diff, dict):
        return False
    for k in key:
        if k in diff:
            return True
    return len(key) == 0


def _simple_merge_diff(result, diff):
    """
    更新业务数据
    :param result: 更新结果
    :param diff: diff pack
    :return:
    """
    for key in list(diff.keys()):
        if diff[key] is None:
            result.pop(key, None)
        elif isinstance(diff[key], dict):
            target = result.setdefault(key, {})
            _simple_merge_diff(target, diff[key])
        else:
            result[key] = diff[key]


def _simple_merge_diff_and_collect_paths(result, diff, path: Tuple, diff_paths: Set, prototype: Union[Dict, None]):
    """
    更新业务数据并收集指定节点的路径
    默认行为 reduce_diff=False，表示函数运行过程中不会修改 diff 本身
    :param result: 更新结果
    :param diff: diff pack
    :param path: 当前迭代 merge_diff 的节点路径
    :param diff_paths: 收集指定节点的路径
    :param prototype: 数据原型, 为 None 的节点路径会被记录在 diff_paths 集合中
    :return:
    """
    for key in list(diff.keys()):
        if diff[key] is None:
            result.pop(key, None)
            if prototype and ('*' in prototype or key in prototype) and prototype['*' if '*' in prototype else key] is None:
                diff_paths.add(path + (key, ))
        elif isinstance(diff[key], dict):
            target = result.setdefault(key, {})
            sub_path = path + (key, )
            sub_prototype = None
            if prototype and ('*' in prototype or key in prototype):
                sub_prototype = prototype['*' if '*' in prototype else key]
                if sub_prototype is None:
                    diff_paths.add(sub_path)
            _simple_merge_diff_and_collect_paths(target, diff[key], path=sub_path, prototype=sub_prototype, diff_paths=diff_paths)
        elif key in result and result[key] == diff[key]:
            pass
        else:
            result[key] = diff[key]
            # 只有确实有变更的字段，会出现在 diff_paths 里
            if prototype and ('*' in prototype or key in prototype) and prototype['*' if '*' in prototype else key] is None:
                diff_paths.add(path + (key, ))



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
__author__ = 'mayanqiong'


from tqsdk.lib.target_pos_task import TargetPosTask, InsertOrderUntilAllTradedTask, InsertOrderTask
from tqsdk.lib.target_pos_scheduler import TargetPosScheduler
from tqsdk.lib.notify import TqNotify
